package gridsim.gssim;

import gridsim.ResourceCharacteristics;
import gridsim.gssim.resource.ResourceProcessors;
import gssim.schedframe.resources.profile.ComputingNodePowerProfile;
import gssim.schedframe.scheduling.ExecTaskInterface;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import schedframe.resources.ComputingNode;
import schedframe.resources.PowerInterface;
import schedframe.resources.ExecutingResourceDescription;
import schedframe.resources.units.Memory;
import schedframe.resources.units.Processor;
import schedframe.resources.units.Processors;
import schedframe.resources.units.ResourceUnit;
import schedframe.resources.units.ResourceUnitType;
import schedframe.scheduling.Reservation;
import schedframe.scheduling.plugin.local.ResourceAllocationInterface;
import schedframe.scheduling.plugin.local.ResourceUnitsManagerInterface;
import schedframe.scheduling.utils.ResourceParameterName;


/**
 * This class represents resource characteristics of a resource
 * 
 * @author Marcin Krystek
 * 
 */

public class ResourceUnitsManagerImpl extends ResourceCharacteristics implements ResourceUnitsManagerInterface, ResourceAllocationInterface {

	public static final int NOT_ADVANCE_RESERVATION = ADVANCE_RESERVATION + 1;
	
	private Log log = LogFactory.getLog(ResourceUnitsManagerImpl.class);

	protected ResourceProcessors processors;

	protected Memory memory;

	protected Map<String, ComputingNode> nodeMap;
	
	protected Map<String, Map<ResourceParameterName, ResourceUnit>> reservedResources; 

	protected Map<String, PowerInterface> nodePowerProfiles;

	/**
	 * Creates new characteristic. resConfig parameter is a map that
	 * contains resource units in a following manner:
	 * <ul>
	 * <li>memory is an Integer stored with a {@link ResourceParameterName#MEMORY}key
	 * <li>Processing elements are stored with a {@link link ResourceParameterName#CPUCOUNT}key
	 * <li>cost per sec is a Double stored with a {@link link ResourceParameterName#COST} key
	 * </ul>
	 * 
	 * @param resConfig
	 *            resource units configuration
	 * @param timeZone
	 *            timezone of resource
	 * @param resReqMatcher
	 *            resource units manager
	 */
	public ResourceUnitsManagerImpl(double timeZone, ExecutingResourceDescription execResDesc){
		super(null, null, null, -1, timeZone, 0);
		
		Collection<ResourceUnit> resConfig = execResDesc.getResourceUnit();
		
		this.nodePowerProfiles = execResDesc.getNodePowerProfiles();
		this.reservedResources = new HashMap<String, Map<ResourceParameterName, ResourceUnit>>();
		this.nodeMap = null;
		
		Iterator<ResourceUnit> itr = resConfig.iterator();
		while(itr.hasNext()){
			ResourceUnit resUnit = itr.next();
			
			switch(resUnit.getName()){
				case COST: this.costPerSec_ = resUnit.getAmount();
					break;
					
				case MEMORY: if(resUnit instanceof Memory){
								this.memory = (Memory) resUnit;
							}
					break;
					
				case CPUCOUNT: if(resUnit instanceof ResourceProcessors){
								this.processors = (ResourceProcessors) resUnit;
							}
					break;
			}
		}
		
	}
	
	/**
	 * Retrieves free resource units that satisfy given resource units
	 * requirements.
	 * 
	 * @param resReq resource units requirements
	 * @return free resource units or null
	 */
	public Map<ResourceParameterName, ResourceUnit> chooseResourcesFor(ExecTaskInterface task) {
		
		if(!(task instanceof SubmittedTask)){
			return null;
		}
		
		// check if there are resources previously reserved for this task.
		String jt = task.getJobId() + "_" + task.getId();
		Map<ResourceParameterName, ResourceUnit> rr = this.reservedResources.get(jt);
		if(rr != null){
			this.reservedResources.remove(jt);
			return rr; 
		}
		
		HashMap<ResourceParameterName, ResourceUnit> map = 
						new HashMap<ResourceParameterName, ResourceUnit>();

		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			cpuRequest = 1;
		}
		
		if(cpuRequest != 0){
			
			if(processors.getFreeAmount() < cpuRequest) {
				log.error("Task requires more cpus than is availiable in this moment.");
				return null;
			}
			
			
			if(task.expectSpecificResource(ResourceParameterName.FREECPUS)){
				String names[] = (String[])task.getExpectedSpecificResource(ResourceParameterName.FREECPUS);
				int cpuIds[] = new int[names.length];
				for(int i = 0; i < names.length; i++){
					cpuIds[i] = Integer.parseInt(names[i]);
				}
				
				ResourceProcessors p = this.processors.clone();
				p.reset();
				for(int i = 0; i < processors.getAmount() && cpuRequest > 0; i++){
					Processor cpu = processors.get(i);
					boolean isChosen = false;
					for(int j = 0; j < cpuIds.length && !isChosen; j++){
						if(cpu.getId() == cpuIds[j])
							isChosen = true;
					}
					if(isChosen && cpu.getStatus() == Processor.Status.FREE){
						p.get(i).setStatus(Processor.Status.BUSY);
						cpuRequest--;
					}
				}
				
				if(cpuRequest > 0)
					return null;
				
				map.put(ResourceParameterName.CPUCOUNT, p);
				
			} else {
			
			//	ResourceProcessors p = this.processors.clone();
				//p.reset();
				
				for(int i = 0; i < processors.getAmount() && cpuRequest > 0; i++){
					if(processors.get(i).getStatus() == Processor.Status.FREE){
						processors.get(i).setStatus(Processor.Status.BUSY);
						cpuRequest--;
					}
				}
			
				map.put(ResourceParameterName.CPUCOUNT, processors);
			}
		}
		
		
		int memoryRequest;
		try {
			memoryRequest = Double.valueOf(task.getMemoryRequest()).intValue();
		} catch (NoSuchFieldException e) {
			memoryRequest = 0;
		}
		if(memoryRequest != 0 && isMemRequirementSatisfied(memoryRequest)){
			Memory m = new Memory(this.memory);
			m.setUsedAmount(memoryRequest);
			map.put(ResourceParameterName.MEMORY, m);
		}
			
		return map;
		
	}

	/**
	 * Checks if given resource units requirements can ever be satisfied by
	 * resource.
	 * 
	 * @param resReq
	 *            resource units requirements
	 * @return true if these resource units can be satisfied, false otherwise
	 */
	public boolean isReqSatisfied(Map<ResourceParameterName, ResourceUnit> resReq) {

		Processors p = (Processors) resReq.get(ResourceParameterName.CPUCOUNT);
		if(p != null){
			if(processors.getProcessorSpeed() < p.getProcessorSpeed())
				return false;
			if(processors.getAmount() < p.getUsedAmount())
				return false;
		}
		
		Memory m = (Memory) resReq.get(ResourceParameterName.MEMORY);
		if(m != null){
			if(memory.getAmount() < m.getUsedAmount())
				return false;
		}
		
		return true;
	}

	/**
	 * 
	 * @return number of PE this resource consists of.
	 */
	public int getNumPE() {
		return processors.getAmount();
	}

	/**
	 * Returns speed of single (first) processing element.
	 */
	public int getPESpeed() {
		return processors.getProcessorSpeed();
	}
	
	/**
	 * Allocates free resource units. This function delegates processing to
	 * {@link gridsim.gssim.interfaces.ResUnitsManagerInterface#allocateResources(Map)}
	 * 
	 * @param freeRes
	 *            resource units
	 */
	public void allocateResources(Map<ResourceParameterName, ResourceUnit> resources) {
		ResourceUnit unit = resources.get(ResourceParameterName.CPUCOUNT);
		
		if(unit != null) {
			
			if(unit instanceof ResourceProcessors){
				ResourceProcessors peUnit = (ResourceProcessors) unit;
				
				for(int i = 0; i < peUnit.getAmount(); i++){
					if(peUnit.get(i).getStatus() == Processor.Status.BUSY){
						//Processor p = processors.get(i);
						//p.setStatus(Processor.Status.BUSY);
					}
				}
			} 
		}

		unit = resources.get(ResourceParameterName.MEMORY);
		if(unit != null)
			memory.setUsedAmount(memory.getUsedAmount() + unit.getUsedAmount());
		
	}

	/**
	 * Free used resources.
	 * 
	 * @param lastUsedResources
	 *            resource units
	 */
	public void freeResources(Map<ResourceParameterName, ResourceUnit> lastUsedResources) {
		ResourceUnit unit = lastUsedResources.get(ResourceParameterName.CPUCOUNT);
		
		if(unit instanceof ResourceProcessors){
			ResourceProcessors peUnit = (ResourceProcessors) unit;
			
			for(int i = 0; i < peUnit.getAmount(); i++){
				if(peUnit.get(i).getStatus() == Processor.Status.BUSY){
					processors.get(i).setStatus(Processor.Status.FREE);
				}
			}
		}

		unit = lastUsedResources.get(ResourceParameterName.MEMORY);
		if(unit != null){
			memory.setUsedAmount(memory.getUsedAmount() - unit.getUsedAmount());
		}
		
	}
	
	public Processors chooseProcessors(Processors cpus){
		ResourceProcessors p = this.processors.clone();
		p.reset();
		int cpuRequest = cpus.getUsedAmount();

		if(cpus instanceof List){
			log.error("CHOOSING SPECIFIC CPUS IN ADVANCE RESERVATION MODEL IS NOT SUPPOERTED YET.");
		} else {
		
			for(int i = 0; i < processors.getAmount() && cpuRequest > 0; i++){
				if(processors.get(i).getStatus() == Processor.Status.FREE){
					p.get(i).setStatus(Processor.Status.BUSY);
					cpuRequest--;
				}
			}
		}
		return p;
	}
	
	public void createResourceReservation(Reservation reservation, Processors processors){
		try {
			ResourceProcessors cpus = (ResourceProcessors) processors;
			for(int i = 0; i < cpus.size(); i++){
				Processor p = cpus.get(i);
				Processor local_p = this.processors.get(i);
				if(p.getStatus() == Processor.Status.BUSY && 
					local_p.getStatus() == Processor.Status.FREE){
					local_p.setStatus(Processor.Status.RESERVED);
				}
			}
			
			HashMap<ResourceParameterName, ResourceUnit> map = 
				new HashMap<ResourceParameterName, ResourceUnit>();
			map.put(ResourceParameterName.CPUCOUNT, processors);
			
			this.reservedResources.put(reservation.getJobId()+"_"+reservation.getTaskId(), map);
			
			}catch(Exception e){
				e.printStackTrace();
			}
	}
	
	public void createResourceReservation(ExecTaskInterface task, Map<ResourceParameterName, ResourceUnit> resources){
		try {
		ResourceProcessors cpus = (ResourceProcessors) resources.get(ResourceParameterName.CPUCOUNT);
		for(int i = 0; i < cpus.size(); i++){
			Processor p = cpus.get(i);
			Processor local_p = this.processors.get(i);
			if(p.getStatus() == Processor.Status.BUSY && 
				local_p.getStatus() == Processor.Status.FREE){
				local_p.setStatus(Processor.Status.RESERVED);
			}
		}
		
		this.reservedResources.put(task.getJobId()+"_"+task.getId(), resources);
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	public void removeResourceReservation(Map<ResourceParameterName, ResourceUnit> lastUsedResources){
		ResourceUnit unit = lastUsedResources.get(ResourceParameterName.CPUCOUNT);
		
		if(unit instanceof ResourceProcessors){
			ResourceProcessors peUnit = (ResourceProcessors) unit;
			
			for(int i = 0; i < peUnit.getAmount(); i++){
				Processor.Status cpuStatus = peUnit.get(i).getStatus();
				if(cpuStatus == Processor.Status.BUSY){
					processors.get(i).setStatus(Processor.Status.FREE);
				}
			}
		}
	}
	
	public boolean areResourcesReserved(ExecTaskInterface task){
		return this.reservedResources.containsKey(task.getJobId()+"_"+task.getId());
	}

	
	public boolean areFreeResourcesSufficient(Map<ResourceParameterName, ResourceUnit> resReq){
		ResourceUnit requirement = null;
		
		// check memory requirements
		if(resReq.containsKey(ResourceParameterName.MEMORY)){
			requirement = resReq.get(ResourceParameterName.MEMORY);

			if(memory.getFreeAmount() < requirement.getUsedAmount())
				return false;
		}
		
		// check cpu requirements
		if(resReq.containsKey(ResourceParameterName.CPUCOUNT)){
			requirement = resReq.get(ResourceParameterName.CPUCOUNT);
			
			if(requirement.getResourceUnitType() == ResourceUnitType.DISCRETE_RESOURCE){
				
				List<Processor> peRequirement = (List<Processor>) requirement;
				for(int i = 0; i < peRequirement.size(); i++){
					if(peRequirement.get(i).getStatus() == Processor.Status.BUSY &&
						processors.get(i).getStatus() == Processor.Status.BUSY) {
						return false;
					}
				}
				
			} else {
				
				int amount = requirement.getUsedAmount();
				for(int i = 0; i < processors.getAmount() && amount > 0; i++){
					if(processors.get(i).getStatus() == Processor.Status.FREE)
						amount--;
				}
				if(amount > 0)
					return false;
			}
		}
		
		return true;
	}


	/**
	 * @author Stanislaw Szczepanowski
	 * @return the current available memory amount
	 */
	public int getFreeMemory() {
		return memory.getFreeAmount();
	}
	
	/**
	 * @author Stanislaw Szczepanowski
	 * @return the total available memory amount
	 */
	public int getTotalMemory() {
		return memory.getAmount();
	}

	
	/**
	 * Checks if memory requirements can be satisfied
	 * 
	 * @param memoryReq
	 * @return true if memory requirements can be satisfied,false otherwise
	 */
	protected boolean isMemRequirementSatisfied(int memoryReq) {
		if (memory.getFreeAmount() < memoryReq)
			return false;
		
		return true;
	}
	
	/**
	 * Checks if given processing element requirement can be satisfied
	 * 
	 * @param peReq
	 *            processing element to be checked
	 * @return true if given processing element requirement can be satisfied,
	 *         false otherwise
	 */
	protected boolean isPeRequirementSatisfied(Processors req) {

		if(processors.getProcessorSpeed() < req.getProcessorSpeed())
			return false;
		
		if(processors.getFreeAmount() < req.getUsedAmount()){
			return false;
		}

		return true;
	}

	public Map<ResourceParameterName, ResourceUnit> getResourceParameters(){
		Map<ResourceParameterName, ResourceUnit> map = new HashMap<ResourceParameterName, ResourceUnit>();
		if(this.processors != null)
			map.put(ResourceParameterName.CPUCOUNT, this.processors);
		if(this.memory != null)
		map.put(ResourceParameterName.MEMORY, this.memory);
		
		return map;
	}
	
	public ResourceUnit getRequestedUnitInstance(ResourceParameterName name){
		if(name == ResourceParameterName.CPUCOUNT){
			Processors p = null;
			if(processors.areValuesUnique()){
				p = processors.clone();
				p.reset();
				
			} else {
				p = new Processors(processors.getResourceId(), processors.getAmount(), 0);
			}
			return p;
		}
		
		if(name == ResourceParameterName.MEMORY){
			return new Memory(memory.getResourceId(), 0, 0);
		}
		return null;
	}
	
	public void printProcessorsState(){
		for(int i = 0; i < this.processors.size(); i++){
			System.err.println(this.processors.get(i).getStatus() + " ");
		}
	}

	@Override
	public Memory getMemory() {
		return this.memory;
	}

	@Override
	public ResourceProcessors getProcessors() {
		return this.processors;
	}

	@Override
	public Map<ResourceParameterName, ResourceUnit> getReservedResources(
			String id) {
		return this.reservedResources.get(id);
	}

	@Override
	public void removeResourceReservation(String id) {
		this.reservedResources.remove(id);
	}

	@Override
	public Collection<ComputingNode> getComputingNode() {
		if(this.nodeMap == null)
			createComputingNodes();
		
		return this.nodeMap.values();
	}

	@Override
	public ComputingNode getComputingNode(String id) {
		if(this.nodeMap == null)
			createComputingNodes();
		
		return this.nodeMap.get(id);
	}
		
	protected void createComputingNodes() {
		String nodeId = this.processors.get(0).getComputingNodeId();
		int fromIndex = 0;
		boolean checkEnMap = (this.nodePowerProfiles != null ? true : false); 
		
		this.nodeMap = new HashMap<String, ComputingNode>();
		
		int last = this.processors.size() - 1;
		for(int i = 0; i < this.processors.size(); i++){
			Processor p = this.processors.get(i);
			String currentNodeId = p.getComputingNodeId();
			if(!nodeId.equals(currentNodeId) || i == last){
				if(i == last) i++;
				ComputingNode computingNode = new ComputingNode(nodeId);
				computingNode.setProcessors(this.processors.subList(fromIndex, i));
				if(checkEnMap){
					PowerInterface enProfile = this.nodePowerProfiles.get(nodeId);
					if(enProfile != null)
						computingNode.accept(enProfile);
					else
						computingNode.accept(new ComputingNodePowerProfile());
				} else {
					computingNode.accept(new ComputingNodePowerProfile());
				}
				
				this.nodeMap.put(nodeId, computingNode);
				
				fromIndex = i;
				nodeId = currentNodeId;
			}
			
		}
	}

}
